@@ -16,9 +16,17 @@ pub enum Op {
     Minus,      // -
     Multiply,   // *
     Divide,     // /
+    Comma,      // ,
     Operand(Operand),
 }
 
+#[derive(PartialEq, Debug, Clone)]
+pub enum Keyword {
+    In,
+    Continue,
+    Break,
+}
+
 #[derive(PartialEq, Debug, Clone)]
 pub enum Operand {
     Literal(String),
@@ -26,6 +34,8 @@ pub enum Operand {
     Decimal(f64),
     Array(Vec<Operand>),
     Object(Object),
+    Keyword(Keyword),
+    Function(String),
 }
 
 // Name that will be found in the execution context.
@@ -35,7 +45,7 @@ pub struct Object {
 }
 
 #[derive(PartialEq, Debug)]
-pub struct Expr {
+pub struct Tokens {
     pub ops: Vec<Op>,
 }
 
@@ -50,7 +60,7 @@ pub enum Action<T> {
     Do(T),
 }
 
-impl Action<Expr> {
+impl Action<Tokens> {
     pub fn add_op(&mut self, op: Op) {
         match self {
             Action::Show(show) => match show {
@@ -84,7 +94,7 @@ pub enum Show<T> {
 
 #[derive(PartialEq, Debug)]
 pub struct Parser {
-    pub parse_tree: Vec<Action<Expr>>,
+    pub parse_tree: Vec<Action<Tokens>>,
 }
 
 impl Parser {
@@ -111,19 +121,36 @@ impl Parser {
                             let tag_to_parse;
                             match template.chars().nth(tag_start + 2).unwrap() {
                                 '=' => {
-                                    parse_tree.push(Action::Show(Show::ExprEscaped(Expr {
+                                    parse_tree.push(Action::Show(Show::ExprEscaped(Tokens {
                                         ops: Vec::new(),
                                     })));
                                     tag_to_parse = &after_tag[1..tag_end];
                                 }
                                 '-' => {
-                                    parse_tree.push(Action::Show(Show::ExprUnescaped(Expr {
+                                    parse_tree.push(Action::Show(Show::ExprUnescaped(Tokens {
                                         ops: Vec::new(),
                                     })));
                                     tag_to_parse = &after_tag[1..tag_end];
                                 }
+                                '#' => {
+                                    // Ignore the comment.
+                                    template = &after_tag[tag_end + 2..];
+                                    index_offset += tag_end + 2;
+                                    continue;
+                                }
+                                '%' => {
+                                    // Show '%'.
+                                    parse_tree.push(Action::Show(Show::Html {
+                                        start: tag_start + 2,
+                                        end: tag_start + 3,
+                                    }));
+
+                                    template = &after_tag[tag_end + 2..];
+                                    index_offset += tag_end + 2;
+                                    continue;
+                                }
                                 _ => {
-                                    parse_tree.push(Action::Do(Expr { ops: Vec::new() }));
+                                    parse_tree.push(Action::Do(Tokens { ops: Vec::new() }));
                                     tag_to_parse = &after_tag[..tag_end];
                                 }
                             }
@@ -155,7 +182,7 @@ impl Parser {
         Ok(Parser { parse_tree })
     }
 
-    fn parse_tag(template: &str, parse_tree: &mut Vec<Action<Expr>>) -> Result<(), String> {
+    fn parse_tag(template: &str, parse_tree: &mut Vec<Action<Tokens>>) -> Result<(), String> {
         let mut current = 0;
         let mut token_begin = 0;
 
@@ -175,9 +202,11 @@ impl Parser {
             }
 
             // Handle Operators.
-            match template[current..current + 1]
-                .find(&['&', '|', '(', ')', '!', '=', '<', '>', '+', '-', '*', '/'][..])
-            {
+            match template[current..current + 1].find(
+                &[
+                    '&', '|', '(', ')', '!', '=', '<', '>', '+', '-', '*', '/', ',',
+                ][..],
+            ) {
                 Some(_) => {
                     Parser::handle_operand(&template[token_begin..current], parse_tree);
 
@@ -201,7 +230,7 @@ impl Parser {
                 token_begin = current + 1;
             } else if current_char == '{' {
                 Parser::handle_token(&template[token_begin..current], parse_tree);
-                parse_tree.push(Action::Do(Expr { ops: Vec::new() }));
+                parse_tree.push(Action::Do(Tokens { ops: Vec::new() }));
                 token_begin = current + 1;
             } else if current_char == '}' {
                 Parser::handle_token(&template[token_begin..current], parse_tree);
@@ -212,7 +241,7 @@ impl Parser {
                 match parse_tree.pop().unwrap() {
                     Action::Do(expr) => {
                         parse_tree.push(Action::Do(expr));
-                        parse_tree.push(Action::Do(Expr { ops: Vec::new() }));
+                        parse_tree.push(Action::Do(Tokens { ops: Vec::new() }));
                     }
                     Action::Show(show) => match show {
                         Show::Html { start: _, end: _ } => {
@@ -222,13 +251,14 @@ impl Parser {
                             // For <%= ... %> tags, only the last expression will be shown.
                             // In other words, <%= a; b %> is identical to <%= b %>
                             parse_tree.push(Action::Do(expr));
-                            parse_tree
-                                .push(Action::Show(Show::ExprUnescaped(Expr { ops: Vec::new() })))
+                            parse_tree.push(Action::Show(Show::ExprUnescaped(Tokens {
+                                ops: Vec::new(),
+                            })))
                         }
                         Show::ExprEscaped(expr) => {
                             parse_tree.push(Action::Do(expr));
                             parse_tree
-                                .push(Action::Show(Show::ExprEscaped(Expr { ops: Vec::new() })))
+                                .push(Action::Show(Show::ExprEscaped(Tokens { ops: Vec::new() })))
                         }
                     },
                     _ => return Err(format!("Invalid position of ; at {}", template)),
@@ -247,7 +277,7 @@ impl Parser {
     fn handle_string_literal(
         template: &str,
         start: usize,
-        parse_tree: &mut Vec<Action<Expr>>,
+        parse_tree: &mut Vec<Action<Tokens>>,
     ) -> Result<usize, String> {
         assert!(template.chars().nth(start).unwrap() == '"');
         let string_literal_end;
@@ -292,7 +322,7 @@ impl Parser {
     fn handle_operator(
         template: &str,
         start: usize,
-        action: &mut Action<Expr>,
+        action: &mut Action<Tokens>,
     ) -> Result<usize, String> {
         // Check for the binary operators.
         if template.len() >= start + 2 {
@@ -314,6 +344,7 @@ impl Parser {
             '-' => action.add_op(Op::Minus),
             '*' => action.add_op(Op::Multiply),
             '/' => action.add_op(Op::Divide),
+            ',' => action.add_op(Op::Comma),
             c => {
                 return Err(format!(
                     "Unknown operator at '{}', unknown : {}",
@@ -325,13 +356,13 @@ impl Parser {
         Ok(start + 1)
     }
 
-    fn handle_token(token: &str, parse_tree: &mut Vec<Action<Expr>>) {
+    fn handle_token(token: &str, parse_tree: &mut Vec<Action<Tokens>>) {
         if token == "if" {
-            parse_tree.push(Action::If(Expr { ops: Vec::new() }));
+            parse_tree.push(Action::If(Tokens { ops: Vec::new() }));
         } else if token == "while" {
-            parse_tree.push(Action::While(Expr { ops: Vec::new() }));
+            parse_tree.push(Action::While(Tokens { ops: Vec::new() }));
         } else if token == "for" {
-            parse_tree.push(Action::For(Expr { ops: Vec::new() }));
+            parse_tree.push(Action::For(Tokens { ops: Vec::new() }));
         } else if token == "else" {
             parse_tree.push(Action::Else());
         } else {
@@ -339,7 +370,7 @@ impl Parser {
         }
     }
 
-    fn handle_operand(mut operand: &str, parse_tree: &mut Vec<Action<Expr>>) {
+    fn handle_operand(mut operand: &str, parse_tree: &mut Vec<Action<Tokens>>) {
         if operand.is_empty() {
             return;
         }
@@ -347,7 +378,7 @@ impl Parser {
         operand = operand.trim();
 
         if let Action::End() = parse_tree.last().unwrap() {
-            parse_tree.push(Action::Do(Expr { ops: Vec::new() }));
+            parse_tree.push(Action::Do(Tokens { ops: Vec::new() }));
         }
 
         let action = parse_tree.last_mut().unwrap();
@@ -362,9 +393,20 @@ impl Parser {
                 action.add_op(Op::Operand(Operand::Number(operand.parse().unwrap_or(0))));
             }
         } else {
-            action.add_op(Op::Operand(Operand::Object(Object {
-                name: operand.to_string(),
-            })))
+            match operand {
+                "in" => {
+                    action.add_op(Op::Operand(Operand::Keyword(Keyword::In)));
+                }
+                "break" => {
+                    action.add_op(Op::Operand(Operand::Keyword(Keyword::Break)));
+                }
+                "continue" => {
+                    action.add_op(Op::Operand(Operand::Keyword(Keyword::Continue)));
+                }
+                _ => action.add_op(Op::Operand(Operand::Object(Object {
+                    name: operand.to_string(),
+                }))),
+            }
         }
     }
 }
@@ -428,7 +470,7 @@ fn parse_simple_expr_with_binary() {
     let result = Parser::parse("<% some == 3 %>");
 
     let expected_expr = Parser {
-        parse_tree: vec![Action::Do(Expr {
+        parse_tree: vec![Action::Do(Tokens {
             ops: vec![
                 Op::Operand(Operand::Object(Object {
                     name: "some".to_string(),
@@ -446,7 +488,7 @@ fn parse_simple_expr_with_unary() {
     let result = Parser::parse("<% !some_value %>");
 
     let expected_expr = Parser {
-        parse_tree: vec![Action::Do(Expr {
+        parse_tree: vec![Action::Do(Tokens {
             ops: vec![
                 Op::Not,
                 Op::Operand(Operand::Object(Object {
@@ -464,7 +506,7 @@ fn parse_complex_binary_and_unary() {
     let result = Parser::parse("<% (a + b * c - e / d) * f %>");
 
     let expected_expr = Parser {
-        parse_tree: vec![Action::Do(Expr {
+        parse_tree: vec![Action::Do(Tokens {
             ops: vec![
                 Op::ParenOpen,
                 Op::Operand(Operand::Object(Object {
@@ -502,7 +544,7 @@ fn parse_complex_binary_and_unary() {
 fn parse_simple_literal() {
     let result = Parser::parse(r#"<% some_value == "abc" %>"#);
     let expected_expr = Parser {
-        parse_tree: vec![Action::Do(Expr {
+        parse_tree: vec![Action::Do(Tokens {
             ops: vec![
                 Op::Operand(Operand::Object(Object {
                     name: "some_value".to_string(),
@@ -520,7 +562,7 @@ fn parse_simple_literal() {
 fn parse_literal_with_escape() {
     let result = Parser::parse(r#"<% some_value == "a\"bc" abc %>"#);
     let expected_expr = Parser {
-        parse_tree: vec![Action::Do(Expr {
+        parse_tree: vec![Action::Do(Tokens {
             ops: vec![
                 Op::Operand(Operand::Object(Object {
                     name: "some_value".to_string(),
@@ -541,7 +583,7 @@ fn parse_literal_with_escape() {
 fn parse_complex_expr() {
     let result = Parser::parse("<% !some_value && ((var1 != var2) || some <= val) %>");
     let expected_expr = Parser {
-        parse_tree: vec![Action::Do(Expr {
+        parse_tree: vec![Action::Do(Tokens {
             ops: vec![
                 Op::Not,
                 Op::Operand(Operand::Object(Object {
@@ -578,7 +620,7 @@ fn parse_complex_expr() {
 fn parse_complex_expr2() {
     let result = Parser::parse(r"<% var1 >= var2 && var2 <= var3 || !var3  %>");
     let expected_expr = Parser {
-        parse_tree: vec![Action::Do(Expr {
+        parse_tree: vec![Action::Do(Tokens {
             ops: vec![
                 Op::Operand(Operand::Object(Object {
                     name: "var1".to_string(),
@@ -612,8 +654,8 @@ fn parse_if_statement() {
     let result = Parser::parse(r#"<% if var1 >= var2 && var3 < "}" { "hello" } %>"#);
     let expected_expr = Parser {
         parse_tree: vec![
-            Action::Do(Expr { ops: vec![] }),
-            Action::If(Expr {
+            Action::Do(Tokens { ops: vec![] }),
+            Action::If(Tokens {
                 ops: vec![
                     Op::Operand(Operand::Object(Object {
                         name: "var1".to_string(),
@@ -630,7 +672,7 @@ fn parse_if_statement() {
                     Op::Operand(Operand::Literal("}".to_string())),
                 ],
             }),
-            Action::Do(Expr {
+            Action::Do(Tokens {
                 ops: vec![Op::Operand(Operand::Literal("hello".to_string()))],
             }),
             Action::End(),
@@ -645,8 +687,8 @@ fn parse_multiple_if_statement() {
     let result = Parser::parse(r#"<% if var1>=var2{if var2>var3{} if !var1 {}} %>"#);
     let expected_expr = Parser {
         parse_tree: vec![
-            Action::Do(Expr { ops: vec![] }),
-            Action::If(Expr {
+            Action::Do(Tokens { ops: vec![] }),
+            Action::If(Tokens {
                 ops: vec![
                     Op::Operand(Operand::Object(Object {
                         name: "var1".to_string(),
@@ -657,8 +699,8 @@ fn parse_multiple_if_statement() {
                     })),
                 ],
             }),
-            Action::Do(Expr { ops: vec![] }),
-            Action::If(Expr {
+            Action::Do(Tokens { ops: vec![] }),
+            Action::If(Tokens {
                 ops: vec![
                     Op::Operand(Operand::Object(Object {
                         name: "var2".to_string(),
@@ -669,9 +711,9 @@ fn parse_multiple_if_statement() {
                     })),
                 ],
             }),
-            Action::Do(Expr { ops: vec![] }),
+            Action::Do(Tokens { ops: vec![] }),
             Action::End(),
-            Action::If(Expr {
+            Action::If(Tokens {
                 ops: vec![
                     Op::Not,
                     Op::Operand(Operand::Object(Object {
@@ -679,7 +721,7 @@ fn parse_multiple_if_statement() {
                     })),
                 ],
             }),
-            Action::Do(Expr { ops: vec![] }),
+            Action::Do(Tokens { ops: vec![] }),
             Action::End(),
             Action::End(),
         ],
@@ -693,8 +735,8 @@ fn parse_if_else_statement() {
     let result = Parser::parse(r#"<% if var1>=var2{ "b" } else if var1==3{"a"}else{"c"}%>"#);
     let expected_expr = Parser {
         parse_tree: vec![
-            Action::Do(Expr { ops: vec![] }),
-            Action::If(Expr {
+            Action::Do(Tokens { ops: vec![] }),
+            Action::If(Tokens {
                 ops: vec![
                     Op::Operand(Operand::Object(Object {
                         name: "var1".to_string(),
@@ -705,12 +747,12 @@ fn parse_if_else_statement() {
                     })),
                 ],
             }),
-            Action::Do(Expr {
+            Action::Do(Tokens {
                 ops: vec![Op::Operand(Operand::Literal("b".to_string()))],
             }),
             Action::End(),
             Action::Else(),
-            Action::If(Expr {
+            Action::If(Tokens {
                 ops: vec![
                     Op::Operand(Operand::Object(Object {
                         name: "var1".to_string(),
@@ -719,12 +761,12 @@ fn parse_if_else_statement() {
                     Op::Operand(Operand::Number(3)),
                 ],
             }),
-            Action::Do(Expr {
+            Action::Do(Tokens {
                 ops: vec![Op::Operand(Operand::Literal("a".to_string()))],
             }),
             Action::End(),
             Action::Else(),
-            Action::Do(Expr {
+            Action::Do(Tokens {
                 ops: vec![Op::Operand(Operand::Literal("c".to_string()))],
             }),
             Action::End(),
@@ -739,8 +781,8 @@ fn parse_while_statement() {
     let result = Parser::parse(r#"<% while var1 >= var2{ "hello"} %>"#);
     let expected_expr = Parser {
         parse_tree: vec![
-            Action::Do(Expr { ops: vec![] }),
-            Action::While(Expr {
+            Action::Do(Tokens { ops: vec![] }),
+            Action::While(Tokens {
                 ops: vec![
                     Op::Operand(Operand::Object(Object {
                         name: "var1".to_string(),
@@ -751,7 +793,7 @@ fn parse_while_statement() {
                     })),
                 ],
             }),
-            Action::Do(Expr {
+            Action::Do(Tokens {
                 ops: vec![Op::Operand(Operand::Literal("hello".to_string()))],
             }),
             Action::End(),
@@ -766,21 +808,19 @@ fn parse_for_statement() {
     let result = Parser::parse(r#"<% for a in vec{ "hello"} %>"#);
     let expected_expr = Parser {
         parse_tree: vec![
-            Action::Do(Expr { ops: vec![] }),
-            Action::For(Expr {
+            Action::Do(Tokens { ops: vec![] }),
+            Action::For(Tokens {
                 ops: vec![
                     Op::Operand(Operand::Object(Object {
                         name: "a".to_string(),
                     })),
-                    Op::Operand(Operand::Object(Object {
-                        name: "in".to_string(),
-                    })),
+                    Op::Operand(Operand::Keyword(Keyword::In)),
                     Op::Operand(Operand::Object(Object {
                         name: "vec".to_string(),
                     })),
                 ],
             }),
-            Action::Do(Expr {
+            Action::Do(Tokens {
                 ops: vec![Op::Operand(Operand::Literal("hello".to_string()))],
             }),
             Action::End(),
@@ -790,6 +830,39 @@ fn parse_for_statement() {
     assert_eq!(result.unwrap(), expected_expr);
 }
 
+#[test]
+fn parse_function() {
+    let result = Parser::parse(r#"<% func(1+1, a, b) + x %>"#);
+    let expected_expr = Parser {
+        parse_tree: vec![Action::Do(Tokens {
+            ops: vec![
+                Op::Operand(Operand::Object(Object {
+                    name: "func".to_string(),
+                })),
+                Op::ParenOpen,
+                Op::Operand(Operand::Number(1)),
+                Op::Plus,
+                Op::Operand(Operand::Number(1)),
+                Op::Comma,
+                Op::Operand(Operand::Object(Object {
+                    name: "a".to_string(),
+                })),
+                Op::Comma,
+                Op::Operand(Operand::Object(Object {
+                    name: "b".to_string(),
+                })),
+                Op::ParenClose,
+                Op::Plus,
+                Op::Operand(Operand::Object(Object {
+                    name: "x".to_string(),
+                })),
+            ],
+        })],
+    };
+
+    assert_eq!(result.unwrap(), expected_expr);
+}
+
 #[test]
 fn parse_tags_simple() {
     let template = r#"<html><% if a { %>some tag<% } %><body></body></html>"#;
@@ -797,15 +870,15 @@ fn parse_tags_simple() {
     let expected_expr = Parser {
         parse_tree: vec![
             Action::Show(Show::Html { start: 0, end: 6 }),
-            Action::Do(Expr { ops: vec![] }),
-            Action::If(Expr {
+            Action::Do(Tokens { ops: vec![] }),
+            Action::If(Tokens {
                 ops: vec![Op::Operand(Operand::Object(Object {
                     name: "a".to_string(),
                 }))],
             }),
-            Action::Do(Expr { ops: vec![] }),
+            Action::Do(Tokens { ops: vec![] }),
             Action::Show(Show::Html { start: 18, end: 26 }),
-            Action::Do(Expr { ops: vec![] }),
+            Action::Do(Tokens { ops: vec![] }),
             Action::End(),
             Action::Show(Show::Html { start: 33, end: 53 }),
         ],
@@ -822,19 +895,19 @@ fn parse_tags_escaped() {
     let expected_expr = Parser {
         parse_tree: vec![
             Action::Show(Show::Html { start: 0, end: 6 }),
-            Action::Do(Expr { ops: vec![] }),
-            Action::If(Expr {
+            Action::Do(Tokens { ops: vec![] }),
+            Action::If(Tokens {
                 ops: vec![Op::Operand(Operand::Object(Object {
                     name: "a".to_string(),
                 }))],
             }),
-            Action::Do(Expr { ops: vec![] }),
-            Action::Show(Show::ExprEscaped(Expr {
+            Action::Do(Tokens { ops: vec![] }),
+            Action::Show(Show::ExprEscaped(Tokens {
                 ops: vec![Op::Operand(Operand::Object(Object {
                     name: "data".to_string(),
                 }))],
             })),
-            Action::Do(Expr { ops: vec![] }),
+            Action::Do(Tokens { ops: vec![] }),
             Action::End(),
         ],
     };
@@ -848,37 +921,37 @@ fn parse_multipe_expressions() {
     let expected_expr = Parser {
         parse_tree: vec![
             Action::Show(Show::Html { start: 0, end: 6 }),
-            Action::Do(Expr {
+            Action::Do(Tokens {
                 ops: vec![Op::Operand(Operand::Object(Object {
                     name: "a".to_string(),
                 }))],
             }),
-            Action::Do(Expr {
+            Action::Do(Tokens {
                 ops: vec![Op::Operand(Operand::Object(Object {
                     name: "b".to_string(),
                 }))],
             }),
-            Action::Do(Expr {
+            Action::Do(Tokens {
                 ops: vec![Op::Operand(Operand::Object(Object {
                     name: "c".to_string(),
                 }))],
             }),
-            Action::Do(Expr {
+            Action::Do(Tokens {
                 ops: vec![Op::Operand(Operand::Object(Object {
                     name: "a".to_string(),
                 }))],
             }),
-            Action::Show(Show::ExprEscaped(Expr {
+            Action::Show(Show::ExprEscaped(Tokens {
                 ops: vec![Op::Operand(Operand::Object(Object {
                     name: "b".to_string(),
                 }))],
             })),
-            Action::Do(Expr {
+            Action::Do(Tokens {
                 ops: vec![Op::Operand(Operand::Object(Object {
                     name: "a".to_string(),
                 }))],
             }),
-            Action::Show(Show::ExprUnescaped(Expr {
+            Action::Show(Show::ExprUnescaped(Tokens {
                 ops: vec![Op::Operand(Operand::Object(Object {
                     name: "b".to_string(),
                 }))],
@@ -888,3 +961,30 @@ fn parse_multipe_expressions() {
 
     assert_eq!(result.unwrap(), expected_expr);
 }
+
+#[test]
+fn parse_ignore_comment() {
+    let result = Parser::parse(r#"<html><%# some comment %></html>"#);
+    let expected_expr = Parser {
+        parse_tree: vec![
+            Action::Show(Show::Html { start: 0, end: 6 }),
+            Action::Show(Show::Html { start: 25, end: 32 }),
+        ],
+    };
+
+    assert_eq!(result.unwrap(), expected_expr);
+}
+
+#[test]
+fn parse_percent_tag() {
+    let result = Parser::parse(r#"<html><%% %></html>"#);
+    let expected_expr = Parser {
+        parse_tree: vec![
+            Action::Show(Show::Html { start: 0, end: 6 }),
+            Action::Show(Show::Html { start: 8, end: 9 }),
+            Action::Show(Show::Html { start: 12, end: 19 }),
+        ],
+    };
+
+    assert_eq!(result.unwrap(), expected_expr);
+}
